!===============================================================
! This program is meant to implement DRLSE (Distance Regularized Level Set Evolution). It was orignally written in MATLAB by Chunming Li (lchunming@gmail.com, i_chunming@hotmail.com ), and converted into fortran (and modified) by Taylor Dahlke (Stanford University) in July, 2013.

! NOTE: If alpha is > 0, then the level set boundary will tend to shrink inward. If alpha is < 0, then the level set boundary will tend to expand outward. You want to choose this parameter based on whether or not your initial level set curve encompasses the area that you want to segment (choose alpha > 0), or whether you have the initial level set curve inside of the body that you wish to segment (choose alpha < 0). 

! This implementation is based upon the following paper:
! C. Li, C. Xu, C. Gui, M. D. Fox, "Distance Regularized Level Set Evolution and Its Application to Image Segmentation", IEEE Trans. Image Processing, vol. 19 (12), pp.3243-3254, 2010.

!==============================================================

program DRLSE_main
use sep
use drlse_mod
implicit none

!===============================================================
 
integer					:: nr, nc, timestep, iter_t, iter_out, i, iter_refine, percent
real					:: mu, lambda, alpha, epsilon, p
logical					:: double_well
real, dimension(:,:), allocatable	:: image, levelset_in, levelset_out, image_smooth, Ix, Iy, f, g
real, dimension(:,:,:), allocatable	:: output


 call sep_init()
 call from_history("n1", nr)
 call from_history("n2", nc)

! Load / Define Parameters
 call from_param("timestep",timestep,1)
 call from_param("mu",mu,0.2)
 call from_param("iter_t",iter_t,5)
 call from_param("iter_out",iter_out,20)
 call from_param("iter_refine",iter_refine,10)
 call from_param("lambda",lambda,5.0)
 call from_param("alpha",alpha,3.0)  ! If alpha < 0, the LS expands. If alpha > 0, the LS shrinks. 
 call from_param("epsilon",epsilon,1.5)
 call from_param("double_well",double_well,.true.)
mu=mu/timestep

! Allocate the internal arrays / dimensions
allocate(image(nr,nc), levelset_in(nr,nc), levelset_out(nr,nc), image_smooth(nr,nc), Ix(nr,nc), Iy(nr,nc), f(nr,nc), g(nr,nc))
allocate(output(nr,nc,(iter_out+1)))

! Load Input Image and Initial Levelset
 call sep_read(image)
 call sep_read(levelset_in, "initial_curve")

! Create the input image gradient " g "
image_smooth = image
 call drlse_init(nr, nc)
 call gradient(image_smooth, Ix, Iy)
f = (Ix**2 + Iy**2)
g = 1/(1+f)

! Perform the Level Set evolution
output(:,:,1) = levelset_in
do i=2,iter_out
    call drlse_edge(levelset_in, g, lambda, mu, alpha, epsilon, timestep, iter_t, double_well,levelset_out)
    levelset_in = levelset_out
    output(:,:,i) = levelset_out
    p = (real(i)/(real(iter_out)+1))*100.0
    percent = int(p)
    write(0,*) " ",percent,"% Complete"  
    write(0,*) " level set test value", levelset_out(nr/2, nc/2)
end do


! Refine the zero contour by further level set evolution with alpha=0. This means the area term is not used, and only the distance regularization term and the edge term are evolving the level set curve.

alpha=0
 call drlse_edge(levelset_in, g, lambda, mu, alpha, epsilon, timestep, iter_refine, double_well,levelset_out)
output(:,:,iter_out+1) = levelset_out
write(0,*) " level set test value", levelset_out(nr/2, nc/2)
write(0,*) " Iterations Complete. Writing output "


! Write the final output
 call sep_write(output)
 call to_history("n1", nr)
 call to_history("n2", nc)
 call to_history("n3", iter_out+1)
 call to_history("label1", 'x')
 call to_history("label2", 'z')
 call to_history("label3", 'Levelset Iteration')
 call sep_close()
end program
