!==============================================================

! These are modules that are used for performing DRLSE (Distance Regularized Level Set Evolution). They were orignally written in MATLAB by Chunming Li (lchunming@gmail.com, i_chunming@hotmail.com ), and converted into fortran by Taylor Dahlke (Stanford University).


! This implementation is based upon the following paper:
! C. Li, C. Xu, C. Gui, M. D. Fox, "Distance Regularized Level Set Evolution and Its Application to Image Segmentation", IEEE Trans. Image Processing, vol. 19 (12), pp.3243-3254, 2010.

!==============================================================
module drlse_mod
implicit none
real, private    			:: pi, small_number
integer, private    			:: nr, nc, i, j, k
contains
!==============================================================



subroutine drlse_init(nr_in, nc_in)
integer           			:: nr_in, nc_in
nr = nr_in
nc = nc_in
pi = 3.1415926535897932384626433832795028841971
small_number=0.0000000001
end subroutine





subroutine drlse_edge_fbn(fbn, phi_0, g, lambda, mu, alpha, epsilon, gamma, time_step, num_iter, double_well, phi)
real, dimension(:,:) 			:: phi_0, g, phi, fbn
real, dimension(nr,nc) 			:: vx, vy, phi_x, phi_y, s, Nx, Ny, curvature, distRegTerm, laplac_mat, dirac_phi, areaTerm, edgeTerm, fbnTerm, phi_in, phi_out
integer               			:: time_step, num_iter
real					:: lambda, mu, alpha, epsilon, gamma
logical					:: double_well

vx=0.0; vy=0.0; phi_x=0.0; phi_y=0.0; s=0.0; Nx=0.0; Ny=0.0; curvature=0.0; distRegTerm=0.0; laplac_mat=0.0; dirac_phi=0.0; areaTerm=0.0; edgeTerm=0.0; phi_in=0.0; phi_out=0.0; phi=0.0; 

phi=phi_0
 call gradient(g, vx, vy)

do k=1,num_iter
	phi_in=phi
	call neumann_bound_cond(phi_in,phi_out)
	call gradient(phi_out, phi_x, phi_y)
	s = sqrt( phi_x**2 + phi_y**2)
	Nx = phi_x/(s + small_number)
	Ny = phi_y/(s + small_number)
	call div(Nx,Ny,curvature)
	! Check for what kind of potential function to use (double well or single well)
	if(double_well)then
		!write(0,*) "Using double well"
		call dist_reg_p2(phi_out,distRegTerm)
	else
		!write(0,*) "Using single well"
		call del2(phi_out,laplac_mat)
		distRegTerm = 4*laplac_mat - curvature
	endif
	call dirac(phi_out,epsilon,dirac_phi)
	fbnTerm = dirac_phi * fbn ! Frobenius Norm force
	areaTerm = dirac_phi * g ! Balloon / pressure force
	edgeTerm = dirac_phi * ((vx * Nx) + (vy * Ny)) + dirac_phi * g * curvature
	phi = phi_out + time_step*(mu*distRegTerm + lambda*edgeTerm + alpha*areaTerm + gamma*fbnTerm)
end do
end subroutine






subroutine drlse_edge(phi_0, g, lambda, mu, alpha, epsilon, time_step, num_iter, double_well, phi)
real, dimension(:,:) 			:: phi_0, g, phi
real, dimension(nr,nc) 			:: vx, vy, phi_x, phi_y, s, Nx, Ny, curvature, distRegTerm, laplac_mat, dirac_phi, areaTerm, edgeTerm, phi_in, phi_out
integer               			:: time_step, num_iter
real					:: lambda, mu, alpha, epsilon
logical					:: double_well

vx=0.0; vy=0.0; phi_x=0.0; phi_y=0.0; s=0.0; Nx=0.0; Ny=0.0; curvature=0.0; distRegTerm=0.0; laplac_mat=0.0; dirac_phi=0.0; areaTerm=0.0; edgeTerm=0.0; phi_in=0.0; phi_out=0.0; phi=0.0; 

phi=phi_0
 call gradient(g, vx, vy)

do k=1,num_iter
	phi_in=phi
	call neumann_bound_cond(phi_in,phi_out)
	call gradient(phi_out, phi_x, phi_y)
	s = sqrt( phi_x**2 + phi_y**2)
	Nx = phi_x/(s + small_number)
	Ny = phi_y/(s + small_number)
	call div(Nx,Ny,curvature)
	! Check for what kind of potential function to use (double well or single well)
	if(double_well) then
		!write(0,*) "Using double well"
		call dist_reg_p2(phi_out,distRegTerm)
	end if
	if(.not. double_well) then
		!write(0,*) "Using single well"
		call del2(phi_out,laplac_mat)
		distRegTerm = 4*laplac_mat - curvature
	endif
	call dirac(phi_out,epsilon,dirac_phi)
	areaTerm = dirac_phi * g ! Balloon / pressure force
	edgeTerm = dirac_phi * ((vx * Nx) + (vy * Ny)) + dirac_phi * g * curvature
	phi = phi_out + time_step*(mu*distRegTerm + lambda*edgeTerm + alpha*areaTerm)
end do
end subroutine










subroutine dist_reg_p2(phi,out)
real, dimension(:,:)			:: phi, out
real, dimension (nr,nc)			:: phi_x, phi_y, a, b, c, d, e, f, s, ps, dps, div_inx, div_iny, lap, temp
 phi_x=0.0; phi_y=0.0; a=0.0; b=0.0; c=0.0; d=0.0; e=0.0; f=0.0; s=0.0; ps=0.0; dps=0.0; div_inx=0.0; div_iny=0.0; lap=0.0; temp=0.0; out=0.0; 

 call gradient(phi,phi_x,phi_y)
s=sqrt(phi_x**2 + phi_y**2)
! Make binary selector matrices
do j=1,nc
    do i=1,nr    
	if(s(i,j) >= 0 .and.  s(i,j) <= 1) then
		a(i,j) = 1   ! "a" matrix
	end if
	if(s(i,j) > 1 ) then
		b(i,j) = 1   ! "b" matrix
	end if
	if(s(i,j) == 0 ) then
		c(i,j) = 1   ! "c" matrix
		else
		d(i,j) = 1   ! "d" matrix
	end if
		ps(i,j)=a(i,j)*sin(2*pi*s(i,j))/(2*pi)+b(i,j)*(s(i,j)-1)
	if(ps(i,j) == 0 ) then
		e(i,j) = 1   ! "e" matrix
		else
		f(i,j) = 1   ! "f" matrix
	end if
	dps(i,j)=(f(i,j)*ps(i,j) + e(i,j))/(d(i,j)*s(i,j) + c(i,j))
    end do
end do
div_inx = (dps*phi_x - phi_x)
div_iny = (dps*phi_y - phi_y) 
 call div(div_inx,div_iny,temp)
 call del2(phi, lap)
out =  temp + 4*lap ! The lap term is what doesnt match the MATLAB code this is based on. This is because a different Laplacian operator is being used. temp is fine though, no problems there.
end subroutine






subroutine gradient(nt_in, nt_dx, nt_dy)
real, dimension(:,:)			:: nt_in, nt_dx, nt_dy
nt_dx=0.0
nt_dy=0.0

do j=1,nc 
    do i=1,nr
	! Middle gradient values
	if(j/=1 .or. j/=nc) then
	nt_dx(i,j) = (nt_in(i,j+1) - nt_in(i,j-1) )/2
	endif
	if(i/=1 .or. i/=nr) then
	nt_dy(i,j) = (nt_in(i+1,j) - nt_in(i-1,j) )/2
	endif
	! Edge dx gradient values
	if(j==1) then
	nt_dx(i,1) = nt_in(i,2) - nt_in(i,1)
	endif
	if(j==nc) then
	nt_dx(i,nc) = nt_in(i,nc) - nt_in(i,nc-1)
	endif
    end do
    ! Edge dy gradient values
    nt_dy(1,j) = nt_in(2,j) - nt_in(1,j)
    nt_dy(nr,j) = nt_in(nr,j) - nt_in(nr-1,j)
end do
end subroutine





subroutine del2(in, out)
real, dimension(:,:) 			:: in, out
real, dimension(nr,nc) 			:: interior, left, right, top, bottom, ul_corner, br_corner 
integer               			:: i, j
real					:: h, ul, ur, bl, br
! Zero out internal arrays
out = 0.0; interior = 0.0; left = 0.0;  right = 0.0;  top = 0.0;  bottom = 0.0;  ul_corner = 0.0;  br_corner = 0.0;
h=2.0

! Interior Points
do j=1,nc
    do i=1,nr
	! Interior Point Calculations
	if( j>1 .and. j<nc .and. i>1 .and. i<nr )then
	interior(i,j) = ((in(i-1,j) + in(i+1,j) + in(i,j-1) + in(i,j+1)) - 4*in(i,j) )/(h**2)
	end if
	! Boundary Conditions for Left and Right edges
	left(i,1) = (-5.0*in(i,2) + 4.0*in(i,3) - in(i,4) + 2.0*in(i,1) + in(i+1,1) + in(i-1,1) - 2.0*in(i,1) )/(h**2)
	right(i,nc) = (-5.0*in(i,nc-1) + 4.0*in(i,nc-2) - in(i,nc-3) + 2.0*in(i,nc) + in(i+1,nc) + in(i-1,nc) - 2.0*in(i,nc) )/(h**2)
    end do
    ! Boundary Conditions for Top and Bottom edges
    top(1,j) = (-5.0*in(2,j) + 4.0*in(3,j) - in(4,j) + 2.0*in(1,j) + in(1,j+1) + in(1,j-1) - 2.0*in(1,j) )/(h**2)
    bottom(nr,j) = (-5.0*in(nr-1,j) + 4.0*in(nr-2,j) - in(nr-3,j) + 2.0*in(nr,j) + in(nr,j+1) + in(nr,j-1) - 2.0*in(nr,j) )/(h**2)
end do
out = interior + left + right + top + bottom 
! Calculate BC for the corners
ul = (-5.0*in(1,2) + 4.0*in(1,3) - in(1,4) + 2.0*in(1,1) - 5.0*in(2,1) + 4.0*in(3,1) - in(4,1) + 2.0*in(1,1))/(h**2)
br = (-5.0*in(nr,nc-1) + 4.0*in(nr,nc-2) - in(nr,nc-3) + 2.0*in(nr,nc) - 5.0*in(nr-1,nc) + 4.0*in(nr-2,nc) - in(nr-3,nc) + 2.0*in(nr,nc))/(h**2)
bl = (-5.0*in(nr,2) + 4.0*in(nr,3) - in(nr,4) + 2.0*in(nr,1) - 5.0*in(nr-1,1) + 4.0*in(nr-2,1) - in(nr-3,1) + 2.0*in(nr,1))/(h**2)
ur = (-5.0*in(1,nc-1) + 4.0*in(1,nc-2) - in(1,nc-3) + 2.0*in(1,nc) - 5.0*in(2,nc) + 4.0*in(3,nc) - in(4,nc) + 2.0*in(1,nc))/(h**2)
! Apply BC for the corners
out(1,1)=ul
out(1,nc)=ur
out(nr,1)=bl
out(nr,nc)=br
end subroutine





subroutine div(nx,ny,out)
real, dimension (:,:)  			:: nx, ny, out
real, dimension (nr,nc)			:: outx, outy, junk
outx=0.0; outy=0.0; junk=0.0; out=0.0;
 call gradient(nx,outx,junk)
 call gradient(ny,junk,outy)
out=outx+outy
end subroutine




subroutine neumann_bound_cond(in,out)
real, dimension(:,:)			:: in, out
out=0.0
out=in
out((/ 1, nr /), (/ 1, nc /)) = out((/ 3, nr-2 /), (/ 3, nc-2 /))
out((/ 1, nr /), 2:(nc-1)) = out((/ 3, nr-2 /), 2:(nc-1))
out(2:(nr-1), (/ 1, nc /)) = out(2:(nr-1), (/ 3, nc-2 /))
end subroutine





subroutine dirac(x,sigma,out)
real					:: sigma
real, dimension (:,:)  			:: x, out
real, dimension (nr,nc)			:: a
a=0; out=0.0;
out=(1/(2*sigma))*(1 + cos(pi*x/sigma))
! Make binary selector matrices
do j=1,nc
    do i=1,nr    
	if(x(i,j) <= sigma .and. x(i,j) >= -sigma) then
		a(i,j) = 1 ! "a" matrix
	endif
    end do
end do
out=out*a
end subroutine







subroutine frobnorm(image, radius, bb, rescale, fbn)
!===============================================================================
! Structure Tensor arrangement:
!   | s1  s2 |
!   | s3  s4 |
!===============================================================================

real, dimension(:,:)			:: image, fbn
real, dimension(:,:), allocatable	:: Ix, Iy
real, dimension(nr,nc)			:: Ixo, Iyo
real					:: s1, s2, s3, s4, aa, bb, max, min
integer					:: ii, jj, nn, radius
logical					:: rescale
nn = 1 + (2*radius)
fbn = 0.0
allocate(Ix(nn,nn))
allocate(Iy(nn,nn))

! Get the gradient values from full image
 call gradient(image, Ixo, Iyo)

! Begin loop over each pixel
do j=1,nc
    do i=1,nr
! Define neighborhood around each pixel
	! Interior pixel
	if(  (i > radius .and. i < (nr+1-radius)) .and. (j > radius .and. j < (nc+1-radius))  )then
	    Ix = Ixo((i-radius):(i+radius),(j-radius):(j+radius))
	    Iy = Iyo((i-radius):(i+radius),(j-radius):(j+radius))
	! Top Edge
	else if(  (i <= radius) .and. (j > radius .and. j < (nc+1-radius)) )then
	    Ix = Ixo(1:nn,(j-radius):(j+radius))
	    Iy = Iyo(1:nn,(j-radius):(j+radius))
	! Bottom Edge
	else if(  (i >= (nr+1-radius)) .and. (j > radius .and. j < (nc+1-radius)) )then
	    Ix = Ixo((nr-nn+1):nr,(j-radius):(j+radius))
	    Iy = Iyo((nr-nn+1):nr,(j-radius):(j+radius))
	! Left Edge
	else if(  (j <= radius) .and. (i > radius .and. i < (nr+1-radius)) )then
	    Ix = Ixo((i-radius):(i+radius),1:nn)
	    Iy = Iyo((i-radius):(i+radius),1:nn)
	! Right Edge
	else if(  (j >= (nc+1-radius)) .and. (i > radius .and. i < (nr+1-radius)) )then
	    Ix = Ixo((i-radius):(i+radius),(nc-nn+1):nc)
	    Iy = Iyo((i-radius):(i+radius),(nc-nn+1):nc)

	! UL Corner
	else if ( (i<=radius) .and. (j<=radius) )then
	    Ix = Ixo(1:nn,1:nn)
	    Iy = Iyo(1:nn,1:nn)
	! UR Corner
	else if( (i<=radius) .and. (j>(nc-radius)) )then
	    Ix = Ixo(1:nn,(nc-nn+1):nc)
	    Iy = Iyo(1:nn,(nc-nn+1):nc)
	! BL Corner
	else if( (i>(nr-radius)) .and. (j<=radius) )then
	    Ix = Ixo(nr-nn+1:nr,1:nn)
	    Iy = Iyo(nr-nn+1:nr,1:nn)
	! BR Corner
	else if( (i>(nr-radius)) .and. (j>(nc-radius)) )then
	    Ix = Ixo(nr-nn+1:nr,nc-nn+1:nc)
	    Iy = Iyo(nr-nn+1:nr,nc-nn+1:nc)
	end if

! Create the structure tensor for the 2D case
	s1= 0.0
	s4= 0.0
	do jj=1,nn
   	    do ii=1,nn
		s1 = s1 + Ix(ii,jj)**2
		s4 = s4 + Iy(ii,jj)**2	
	    end do
	end do
! Calc out the frobenius norm (fn)
	fbn(i,j) = s1**2 + s4**2
    end do
end do

! Rescale the frobenius norm
if(rescale)then
    write(0,*) "Frobenius Rescaling Used"
    fbn = log(fbn)
    max=maxval(fbn)
    min=minval(fbn)
    aa=min + bb*(max - min)
    do i=1,nr
	do j=1,nc
	   if( fbn(i,j) < aa) then
		fbn(i,j) = -1
	   else
		fbn(i,j) = 1
	   end if
	end do
    end do
else
    write(0,*) "No Frobenius Rescaling"
endif
end subroutine

end module

